#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include <unistd.h>
#include <netdb.h>

#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/fcntl.h>

int	sock;

static void
ex(const char *s)
{
  perror(s);
  exit(2);
}

static void
getaddr(struct sockaddr_in *sin, const char *host, int port)
{
  memset(sin, 0, sizeof *sin);
  sin->sin_family	= AF_INET;
  sin->sin_addr.s_addr	= htonl(INADDR_ANY);
  sin->sin_port		= htons(port);
  if (host && !inet_aton(host, &sin->sin_addr))
    {
      struct hostent	*he;

      if ((he=gethostbyname(host))==0)
	ex(host);
      if (he->h_addrtype!=AF_INET || he->h_length!=sizeof sin->sin_addr)
	ex("unsupported host address");
      memcpy(&sin->sin_addr, he->h_addr, sizeof sin->sin_addr);
    }
}

static void
sockopen(const char *host, int port)
{
  struct sockaddr_in	sin;
  int			on;
  struct linger		linger;

  sock	= socket( AF_INET, SOCK_STREAM, 0 );
  if (sock<0)
    ex("socket");

  on = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on))
    ex("setsockopt reuse");
  linger.l_onoff	= 1;
  linger.l_linger	= 65535;
  if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger))
    ex("setsockopt linger");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &on, sizeof on))
    ex("setsockopt sndbuf");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &on, sizeof on))
    ex("setsockopt rcvbuf");

  getaddr(&sin, host, port);

  if (bind(sock, (struct sockaddr *)&sin, sizeof sin))
    ex("bind");

#if 0
  if (fcntl(sock, F_SETFL, O_NDELAY))
    ex("O_NDELAY");
#endif
}

static void
docopy(int fi, int fo)
{
  char	buf[102400];
  int	got;

  while ((got=read(fi, buf, sizeof buf))!=0)
    {
      int	pos, put;

      if (got<0)
	{
	  if (errno==EINTR || errno==EAGAIN)
	    continue;
	  ex("input");
	}
      for (pos=0; pos<got; pos+=put)
	if ((put=write(fo, buf+pos, got-pos))<=0)
	  {
	    if (!put)
	      {
		fprintf(stderr, "premature EOF on output\n");
		exit(3);
	      }
	    put	= 0;
	    if (errno==EINTR || errno==EAGAIN)
	      continue;
	    ex("output");
	  }
    }
}

static void
copyloop(int sock)
{
  pid_t	pid, a, b;
  int	st;

  if ((a=fork())==0)
    {
      docopy(sock, fileno(stdout));
      exit(0);
    }
  if ((b=fork())==0)
    {
      docopy(fileno(stdin), sock);
      exit(0);
    }
  st	= 4;
  while ((pid=wait(&st))==(pid_t)-1)
    {
      if (errno==ECHILD)
	break;
      ex("wait");
    }
  if (a!=pid && a!=(pid_t)-1)
    kill(a, SIGTERM);
  if (b!=pid && a!=(pid_t)-1)
    kill(b, SIGTERM);
  close(sock);
  exit(st);
}

static void
do_accept(int port)
{
  struct sockaddr_in	sin;
  int			len;
  int			fd;

  sockopen(NULL, port);

  if (listen(sock, 1))
    ex("listen");

  len	= sizeof sin;
  if ((fd=accept(sock, &sin, &len))<0)
    ex("accept");

  copyloop(fd);
}

static void
do_connect(const char *host, int port)
{
  struct sockaddr_in	sin;

  sockopen(NULL, 0);

  getaddr(&sin, host, port);
  if (connect(sock, &sin, sizeof sin))
    ex("connect");
  
  copyloop(sock);
}

int
main(int argc, char **argv)
{
  switch (argc)
    {
    case 2:
      do_accept(atoi(argv[1]));
      break;

    case 3:
      do_connect(argv[1], atoi(argv[2]));
      break;

    default:
      fprintf(stderr, "Usage: %s [host] port\n", argv[0]);
      return 1;
    }
  return 0;
}
